import numpy as np
from dearning.model import CustomAIModel, Dense, Activation, DOtensor
from dearning.training import train_model
from dearning.utils import evaluate_model
from dearning.AI_tools import RLTools
from dearning import libD as nd
from dearning import (
    train_model, translate, TextToSpeech,
    CodeTracker, BinaryConverter
)

def build_basic_model():
    model = CustomAIModel(loss="cross_entropy")
    model.add(Dense(7, 16))
    model.add(Activation("relu"))
    model.add(Dense(16, 8))
    model.add(Activation("tanh"))
    model.add(Dense(8, 4))
    model.add(Activation("softmax"))
    return model

def full_pipeline(task="classification"):
    print("🚀 Menjalankan pipeline dearning...")
    model = build_basic_model()
    model, evaluation = train_model(model=model, task=task)
    print("🎯 Evaluasi Model:", evaluation)

def rl_pipeline():
    print("🧠 Menjalankan Reinforcement Learning pipeline...")
    rl = RLTools()
    rl.add_q_agent()
    rl.add_random_agent()
    rl.run(episodes=50)
    
def run_pipeline(task="translate", input_data=None):
    if task == "translate":
        print("Terjemahan:", translate(input_data))
    elif task == "tts":
        TextToSpeech(input_data)
    elif task == "track":
        CodeTracker().track(input_data)
    elif task == "binary":
        bincode = BinaryConverter().convert_file(input_data)
        print("Hasil Biner:", bincode)
    
def train_model(model, task="classification", X=None, y=None, 
                epochs=100, learning_rate=0.01, batch_size=32, 
                visualize=True, use_autograd=False, use_dotensor=False, trace=False):
    """
    Pipeline standar untuk pelatihan model.
    """
    if X is None or y is None:
        X = np.random.rand(300, 4)
        y = np.random.randint(0, 2, size=(300, 1) if task == "classification" else (300,))

    if use_dotensor:
        X_tensor = [DOtensor(x, requires_grad=True) for x in X]
        y_tensor = [DOtensor(y_val) for y_val in y]
        
        for epoch in range(epochs):
            total_loss = 0
            for x_t, y_t in zip(X_tensor, y_tensor):
                out = model.forward(x_t.data.reshape(1, -1))
                pred = DOtensor(out[0])
                target = y_t
                loss = (pred - target) * (pred - target)

                if trace:
                    print(f"🧠 Loss graph: {loss}")

                loss.backward()
                total_loss += float(loss.data)

            avg_loss = total_loss / len(X_tensor)
            print(f"[DOtensor] Epoch {epoch+1}/{epochs} - Loss: {avg_loss:.4f}")

        return model, {"loss": avg_loss}
    
    else:
        model.train(X, y, epochs=epochs, learning_rate=learning_rate, batch_size=batch_size, verbose=visualize)
        return model, {"loss": model.losses[-1] if model.losses else None}
    
def run_ai(model_name, task_input):
    return 


import numpy as np
from PIL import Image
import io, os, ast, struct, pyttsx3, wave, logging, contextlib
import soundfile as sf
from scipy.signal import resample
from scipy.fftpack import dct
try:
    from textblob import TextBlob
    TEXTBLOB_AVAILABLE = True
except ImportError:
    TEXTBLOB_AVAILABLE = False

logging.basicConfig(level=logging.INFO)

def ekstract():    
    
    try:
        import networkx as nx
        NETWORKX_AVAILABLE = True
    except ImportError:
        NETWORKX_AVAILABLE = False
    try:
        from geopy.geocoders import Nominatim
        GEOPY_AVAILABLE = True
    except ImportError:
        GEOPY_AVAILABLE = False
    try:
        import serial
        PYSERIAL_AVAILABLE = True
    except ImportError:
        PYSERIAL_AVAILABLE = False

# === 📖 NLP: Analisis ===
class DLP:
    def __init__(self, lang="en"):
        self.lang = lang

    def analyze_sentiment(self, text):
        blob = TextBlob(text)
        return {
            "polarity": blob.sentiment.polarity,
            "subjectivity": blob.sentiment.subjectivity,
            "label": "positive" if blob.sentiment.polarity > 0
                     else "negative" if blob.sentiment.polarity < 0
                     else "neutral"
        }

    def extract_nouns(self, text):
        blob = TextBlob(text)
        return list(blob.noun_phrases)

    def pos_tagging(self, text):
        blob = TextBlob(text)
        return blob.tags  # [('word', 'POS'), ...]

    def summarize(self, text, max_sentences=2):
        sentences = text.split(". ")
        return ". ".join(sentences[:max_sentences]) + ("." if len(sentences) > max_sentences else "")

    def process(self, text):
        result = {
            "sentiment": self.analyze_sentiment(text),
            "nouns": self.extract_nouns(text),
            "pos_tags": self.pos_tagging(text),
            "summary": self.summarize(text)
        }
        return result
        
 # === Reinforcement Learning Tools ===
try:
    from simple_rl.agents import QLearningAgent, RandomAgent
    from simple_rl.tasks import GridWorldMDP
    from simple_rl.run_experiments import run_agents_on_mdp

    class RLTools:
        def __init__(self):
            self.env = GridWorldMDP()
            self.agents = []

        def add_q_agent(self, name="q_agent", alpha=0.1, epsilon=0.1, gamma=0.9):
            agent = QLearningAgent(name=name, actions=self.env.get_actions(),
                                   alpha=alpha, epsilon=epsilon, gamma=gamma)
            self.agents.append(agent)
            return agent

        def add_random_agent(self, name="random"):
            agent = RandomAgent(name=name, actions=self.env.get_actions())
            self.agents.append(agent)
            return agent

        def run(self, episodes=100):
            if self.agents:
                run_agents_on_mdp(self.agents, self.env, instances=1, episodes=episodes)
            else:
                print("[⚠️] Tidak ada agen RL yang ditambahkan.")

except ImportError:
    class RLTools:
        def __init__(self):
            print("[❌] simple_rl belum terpasang. Gunakan: pip install simple_rl")

        def add_q_agent(self, *args, **kwargs): pass
        def add_random_agent(self, *args, **kwargs): pass
        def run(self, *args, **kwargs): pass

# === TEXT TO SPEECH ===
class TextToSpeech:
    def __init__(self, voice=None, rate=150, volume=1.0):
        self.engine = pyttsx3.init()
        self.engine.setProperty('rate', rate)
        self.engine.setProperty('volume', volume)
        if voice:
            self.set_voice(voice)

    def set_voice(self, voice_name):
        voices = self.engine.getProperty('voices')
        for voice in voices:
            if voice_name.lower() in voice.name.lower():
                self.engine.setProperty('voice', voice.id)
                break

    def speak(self, text):
        self.engine.say(text)
        self.engine.runAndWait()

# === MEMORY MANAGEMENT ===
MEMORY_PATH = os.path.join(os.path.dirname(__file__), "..", "Memory", "DATAI.py")
MEMORY_VAR = "memory"

class AImemory:
    def __init__(self):
        self.memory = self._load_memory()

    def _load_memory(self):
        if not os.path.exists(MEMORY_PATH):
            with open(MEMORY_PATH, "w") as f:
                f.write(f"{MEMORY_VAR} = []\n")
            return []

        with open(MEMORY_PATH, "r") as f:
            try:
                content = f.read()
                parsed = ast.parse(content, mode='exec')
                for node in parsed.body:
                    if isinstance(node, ast.Assign) and node.targets[0].id == MEMORY_VAR:
                        return ast.literal_eval(ast.unparse(node.value))
            except Exception as e:
                print("❌ Gagal baca memori:", e)
        return []

    def _save_memory(self):
        with open(MEMORY_PATH, "w") as f:
            f.write(f"{MEMORY_VAR} = {repr(self.memory)}\n")

    def add(self, data):
        if data not in self.memory:
            self.memory.append(data)
            self._save_memory()

    def remove(self, data):
        if data in self.memory:
            self.memory.remove(data)
            self._save_memory()

    def clear(self):
        self.memory = []
        self._save_memory()

    def get_all(self):
        return self.memory

    def contains(self, query):
        return query in self.memory

# === 📷 Gambar ===

def load_image(path, target_size=(64, 64), grayscale=False):
    """
    Membaca dan resize gambar, lalu normalisasi + auto-contrast + clipping.
    """
    mode = "L" if grayscale else "RGB"
    img = Image.open(path).convert(mode).resize(target_size)
    arr = np.asarray(img).astype(np.float32)

    # Normalisasi 0-1 + kontras adaptif
    arr -= arr.min()
    if arr.max() > 0:
        arr /= arr.max()

    arr = np.clip(arr, 0.0, 1.0)
    return arr

def flatten_image(img_array):
    """
    Ubah gambar ke vektor 1D (flatten).
    """
    return img_array.flatten()

# === 🎧 Audio ===

def load_audio(path, sr=22050):
    """Membaca audio & resample ke sample rate target (mono)."""
    data, orig_sr = sf.read(path)
    if data.ndim > 1:  # stereo → ambil channel pertama
        data = data[:, 0]
    if orig_sr != sr:
        n_samples = int(len(data) * sr / orig_sr)
        data = resample(data, n_samples)
    return data.astype(np.float32), sr

def extract_mfcc(path, n_mfcc=13, sr=22050, nfilt=26, nfft=512):
    """Ekstrak MFCC ringan pakai numpy+scipy."""
    signal, sr = load_audio(path, sr)

    # === Pre-emphasis ===
    emphasized = np.empty_like(signal)
    emphasized[0] = signal[0]
    emphasized[1:] = signal[1:] - 0.97 * signal[:-1]

    # === Framing ===
    frame_size, frame_stride = 0.025, 0.01
    frame_length = int(sr * frame_size)
    frame_step = int(sr * frame_stride)
    signal_length = len(emphasized)
    num_frames = 1 + int(np.ceil((signal_length - frame_length) / frame_step))
    pad_length = (num_frames - 1) * frame_step + frame_length
    pad_signal = np.pad(emphasized, (0, pad_length - signal_length), mode='constant')

    # Buat frame view (tanpa copy besar) → hemat memori
    shape = (num_frames, frame_length)
    strides = (pad_signal.strides[0]*frame_step, pad_signal.strides[0])
    frames = np.lib.stride_tricks.as_strided(pad_signal, shape=shape, strides=strides)

    # === Windowing (Hamming) ===
    frames = frames * np.hamming(frame_length)

    # === FFT & Power Spectrum ===
    mag_frames = np.abs(np.fft.rfft(frames, nfft))
    pow_frames = (1.0 / nfft) * (mag_frames ** 2)

    # === Mel Filterbank ===
    low_mel, high_mel = 0, 2595 * np.log10(1 + (sr/2) / 700)
    mel_points = np.linspace(low_mel, high_mel, nfilt + 2)
    hz_points = 700 * (10**(mel_points/2595) - 1)
    bin = np.floor((nfft + 1) * hz_points / sr).astype(int)

    fbank = np.zeros((nfilt, nfft//2 + 1))
    for m in range(1, nfilt+1):
        f_m_minus, f_m, f_m_plus = bin[m-1], bin[m], bin[m+1]
        fbank[m-1, f_m_minus:f_m] = (np.arange(f_m_minus, f_m) - bin[m-1]) / (f_m - f_m_minus)
        fbank[m-1, f_m:f_m_plus] = (bin[m+1] - np.arange(f_m, f_m_plus)) / (bin[m+1] - f_m)

    filter_banks = np.dot(pow_frames, fbank.T)
    filter_banks = np.where(filter_banks > 0, np.log(filter_banks), np.finfo(float).eps)

    # === MFCC via DCT ===
    mfcc = dct(filter_banks, type=2, axis=1, norm='ortho')[:, :n_mfcc]
    return mfcc.astype(np.float32)

def get_audio_duration(path):
    """
    Mendapatkan durasi audio WAV secara cepat tanpa dependencies berat.
    """
    with contextlib.closing(wave.open(path, 'r')) as wf:
        frames = wf.getnframes()
        rate = wf.getframerate()
        duration = frames / float(rate)
        return duration

# === 🎥 Video (GIF Frames) ===

def extract_frames_from_gif(path, max_frames=10):
    """
    Ekstrak beberapa frame dari GIF sebagai array numpy, dengan preprocessing.
    """
    img = Image.open(path)
    frames = []
    try:
        for i in range(max_frames):
            img.seek(i)
            frame = img.convert("RGB")
            arr = np.array(frame).astype(np.float32)

            # Tingkatkan akurasi visual
            arr -= arr.mean()
            std = arr.std() + 1e-5
            arr /= std
            arr = np.clip((arr + 0.5), 0.0, 1.0)  # Dinormalisasi kembali ke 0–1

            frames.append(arr)
    except EOFError:
        pass  # End of GIF

    return np.array(frames)

# === 📈 Booster (Optimisasi array/data) ===

def normalize_array(arr):
    """
    Normalisasi array ke [0, 1]
    """
    arr = np.asarray(arr)
    return (arr - np.min(arr)) / (np.max(arr) - np.min(arr) + 1e-8)

def one_hot_encode(labels, num_classes=None):
    """
    Ubah label integer menjadi one-hot encoding.
    """
    labels = np.array(labels).astype(int)
    if num_classes is None:
        num_classes = np.max(labels) + 1
    one_hot = np.zeros((labels.size, num_classes))
    one_hot[np.arange(labels.size), labels] = 1
    return one_hot

def softmax(x):
    """
    Fungsi softmax untuk output klasifikasi.
    """
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=-1, keepdims=True)

# === 🧠 Pemahaman cepat (analisis ringan) ===

def top_k_probs(preds, k=3):
    """
    Ambil top-k prediksi dari hasil softmax.
    """
    sorted_idx = np.argsort(preds)[::-1]
    top_k = [(int(i), float(preds[i])) for i in sorted_idx[:k]]
    return top_k

def summarize_array(arr):
    """
    Ringkasan array (min, max, mean, shape)
    """
    return {
        "min": float(np.min(arr)),
        "max": float(np.max(arr)),
        "mean": float(np.mean(arr)),
        "shape": arr.shape
    }


import numpy as np
from sklearn.metrics import (
    mean_squared_error,
    r2_score,
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    confusion_matrix,
    classification_report
)
from sklearn.preprocessing import StandardScaler
from joblib import Parallel, delayed
import logging

logging.basicConfig(level=logging.INFO)

_scaler = StandardScaler()

# === 🔧 Preprocessing Otomatis ===
def preprocess_data(data, n_jobs=-1):
    if isinstance(data, list):
        data = np.array(data)

    if data.shape[0] > 1000:
        batch_size = 200
        batches = [data[i:i + batch_size] for i in range(0, len(data), batch_size)]
        scaled_batches = Parallel(n_jobs=n_jobs)(
            delayed(_scaler.fit_transform)(batch) for batch in batches
        )
        return np.vstack(scaled_batches)
    else:
        return _scaler.fit_transform(data)

# === 🧠 Evaluasi Otomatis (Regresi/Klasifikasi) ===
def evaluate_model(model, data, labels=None, task=None, threshold=0.5):
    y_pred = model.forward(data)

    # Tentukan jenis tugas jika tidak diberikan
    if task is None:
        task = "classification" if set(np.unique(labels).flatten()) <= {0, 1} else "regression"
        logging.info(f"[Auto Task Detection] Deteksi tugas: {task}")

    result = {}

    if labels is None:
        return {"output_mean": float(np.mean(y_pred))}

    if task == "regression":
        result["mse"] = float(mean_squared_error(labels, y_pred))
        result["r2"] = float(r2_score(labels, y_pred))
        result["mean_error"] = float(np.mean(np.abs(y_pred - labels)))

    elif task == "classification":
        y_class = (y_pred > threshold).astype(int)
        result["accuracy"] = float(accuracy_score(labels, y_class))
        result["precision"] = float(precision_score(labels, y_class, zero_division=0))
        result["recall"] = float(recall_score(labels, y_class, zero_division=0))
        result["f1_score"] = float(f1_score(labels, y_class, zero_division=0))
        result["confusion_matrix"] = confusion_matrix(labels, y_class).tolist()
        result["report"] = classification_report(labels, y_class, output_dict=True)

    else:
        raise ValueError("task harus 'regression' atau 'classification'")

    return result
 
# optimizer canggih   
class Optimizer:
    def __init__(self, method="sgd", learning_rate=0.01, beta1=0.9, beta2=0.999, epsilon=1e-8):
        self.method = method.lower()
        self.lr = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.m_w, self.v_w = {}, {}
        self.m_b, self.v_b = {}, {}
        self.t = 1

    def update(self, w, b, grad_w, grad_b, layer_idx):
        if self.method == "sgd":
            return w - self.lr * grad_w, b - self.lr * grad_b

        elif self.method == "momentum":
            if layer_idx not in self.m_w:
                self.m_w[layer_idx] = np.zeros_like(grad_w)
                self.m_b[layer_idx] = np.zeros_like(grad_b)
            self.m_w[layer_idx] = self.beta1 * self.m_w[layer_idx] + (1 - self.beta1) * grad_w
            self.m_b[layer_idx] = self.beta1 * self.m_b[layer_idx] + (1 - self.beta1) * grad_b
            return w - self.lr * self.m_w[layer_idx], b - self.lr * self.m_b[layer_idx]

        elif self.method == "rmsprop":
            if layer_idx not in self.v_w:
                self.v_w[layer_idx] = np.zeros_like(grad_w)
                self.v_b[layer_idx] = np.zeros_like(grad_b)
            self.v_w[layer_idx] = self.beta2 * self.v_w[layer_idx] + (1 - self.beta2) * grad_w**2
            self.v_b[layer_idx] = self.beta2 * self.v_b[layer_idx] + (1 - self.beta2) * grad_b**2
            return (w - self.lr * grad_w / (np.sqrt(self.v_w[layer_idx]) + self.epsilon),
                    b - self.lr * grad_b / (np.sqrt(self.v_b[layer_idx]) + self.epsilon))

        elif self.method == "adam":
            if layer_idx not in self.m_w:
                self.m_w[layer_idx] = np.zeros_like(grad_w)
                self.v_w[layer_idx] = np.zeros_like(grad_w)
                self.m_b[layer_idx] = np.zeros_like(grad_b)
                self.v_b[layer_idx] = np.zeros_like(grad_b)

            self.m_w[layer_idx] = self.beta1 * self.m_w[layer_idx] + (1 - self.beta1) * grad_w
            self.v_w[layer_idx] = self.beta2 * self.v_w[layer_idx] + (1 - self.beta2) * grad_w**2
            self.m_b[layer_idx] = self.beta1 * self.m_b[layer_idx] + (1 - self.beta1) * grad_b
            self.v_b[layer_idx] = self.beta2 * self.v_b[layer_idx] + (1 - self.beta2) * grad_b**2

            m_w_hat = self.m_w[layer_idx] / (1 - self.beta1 ** self.t)
            v_w_hat = self.v_w[layer_idx] / (1 - self.beta2 ** self.t)
            m_b_hat = self.m_b[layer_idx] / (1 - self.beta1 ** self.t)
            v_b_hat = self.v_b[layer_idx] / (1 - self.beta2 ** self.t)

            self.t += 1

            w_update = self.lr * m_w_hat / (np.sqrt(v_w_hat) + self.epsilon)
            b_update = self.lr * m_b_hat / (np.sqrt(v_b_hat) + self.epsilon)
            return w - w_update, b - b_update

        else:
            raise ValueError(f"Optimizer '{self.method}' tidak dikenali.")


import numpy as np
import json, os, threading
from scipy.signal import convolve
import networkx as nx

class DOtensor:
    _global_trace_enabled = False
    _global_trace_log = []

    def _init_(self, data, requires_grad=False):
        self.data = np.array(data, dtype=float, copy=False)  # hindari copy berlebih
        self.requires_grad = requires_grad
        self.grad = np.zeros_like(self.data) if requires_grad else None
        self._backward = lambda: None
        self._prev = ()

    def _add_(self, other):
        other = other if isinstance(other, DOtensor) else DOtensor(other)
        out = DOtensor(self.data + other.data, requires_grad=(self.requires_grad or other.requires_grad))

        if out.requires_grad:
            def _backward():
                if self.requires_grad:
                    self.grad += out.grad
                if other.requires_grad:
                    other.grad += out.grad
            out._backward = _backward
            out._prev = (self, other)

        if DOtensor._global_trace_enabled:
            DOtensor._global_trace_log.append(("add", self.data, other.data))
        return out

    def _mul_(self, other):
        other = other if isinstance(other, DOtensor) else DOtensor(other)
        out = DOtensor(self.data * other.data, requires_grad=(self.requires_grad or other.requires_grad))

        if out.requires_grad:
            def _backward():
                if self.requires_grad:
                    self.grad += other.data * out.grad
                if other.requires_grad:
                    other.grad += self.data * out.grad
            out._backward = _backward
            out._prev = (self, other)

        if DOtensor._global_trace_enabled:
            DOtensor._global_trace_log.append(("mul", self.data, other.data))
        return out

    def backward(self):
        if not self.requires_grad:
            raise RuntimeError("backward() called on tensor without requires_grad")

        self.grad = np.ones_like(self.data)
        stack = [self]
        visited = set()

        while stack:
            node = stack.pop()
            if node not in visited:
                visited.add(node)
                node._backward()
                stack.extend(node._prev)

    def _repr_(self):
        return f"DOtensor(data={self.data}, grad={self.grad})"

    @staticmethod
    def enable_trace():
        DOtensor._global_trace_enabled = True
        DOtensor._global_trace_log = []

    @staticmethod
    def disable_trace():
        DOtensor._global_trace_enabled = False

    @staticmethod
    def get_trace_log():
        return list(DOtensor._global_trace_log)  # copy ringan

# === Modular Layer Base ===
class Layer:
    def forward(self, x): raise NotImplementedError
    def backward(self, grad): raise NotImplementedError
    def update(self, lr): pass

# === Dense Layer ===
class Dense(Layer):
    def __init__(self, input_dim, output_dim):
        self.weights = np.random.randn(input_dim, output_dim) * np.sqrt(2. / input_dim)
        self.bias = np.zeros((1, output_dim))

    def forward(self, x):
        self.input = x
        self.output = np.dot(x, self.weights) + self.bias
        return self.output

    def backward(self, grad_output):
        self.grad_w = np.dot(self.input.T, grad_output)
        self.grad_b = np.sum(grad_output, axis=0, keepdims=True)
        return np.dot(grad_output, self.weights.T)

    def update(self, lr):
        self.weights -= lr * self.grad_w
        self.bias -= lr * self.grad_b

# === Activation Layer ===
class Activation(Layer):
    def __init__(self, kind="relu"):
        self.kind = kind

    def forward(self, x):
        self.input = x
        if self.kind == "relu":
            return np.maximum(0, x)
        elif self.kind == "sigmoid":
            return 1 / (1 + np.exp(-x))
        elif self.kind == "tanh":
            return np.tanh(x)
        return x

    def backward(self, grad_output):
        if self.kind == "relu":
            return grad_output * (self.input > 0)
        elif self.kind == "sigmoid":
            s = 1 / (1 + np.exp(-self.input))
            return grad_output * s * (1 - s)
        elif self.kind == "tanh":
            return grad_output * (1 - np.tanh(self.input)**2)
        return grad_output

# === Dropout (Opsional) ===
class Dropout(Layer):
    def __init__(self, rate=0.5):
        self.rate = rate

    def forward(self, x):
        self.mask = (np.random.rand(*x.shape) > self.rate).astype(float)
        return x * self.mask

    def backward(self, grad_output):
        return grad_output * self.mask

# === Custom Model ===
class CustomAIModel:
    def __init__(self, loss="mse"):
        self.layers = []
        self.loss = loss
        self.losses = []
        self.memory_neuron = None
        self.expert_neurons = []

    def add(self, layer):
        self.layers.append(layer)

    def forward(self, x):
        for layer in self.layers:
            x = layer.forward(x)
        return x

    def addneuron(self, kind="memory", **kwargs):
        """
        Menambahkan neuron khusus ke model.

        kind: Jenis neuron yang ditambahkan.
        kwargs: Parameter tambahan spesifik neuron.
        """

        if kind == "memory":
            size = kwargs.get("size", 128)
            self.memory_neuron = np.zeros((1, size))
            print(f"🧠 Neuron Memory ditambahkan (size={size})")

        elif kind == "attention":
            def attention(query, keys, values):
                scores = np.dot(query, keys.T)
                weights = np.exp(scores) / np.sum(np.exp(scores), axis=1, keepdims=True)
                return np.dot(weights, values)
            self.attention = attention
            print("🎯 Neuron Attention ditambahkan")

        elif kind == "ntc":            
            def ntc_layer(x, kernel):
                return convolve(x, kernel, mode='same')
            self.ntc = lambda x: ntc_layer(x, kwargs.get("kernel", np.array([1, -1, 1])))
            print("🕒 Neuron Temporal Convolution ditambahkan")

        elif kind == "graph":            
            self.graph = nx.Graph()
            self.graph.add_nodes_from(kwargs.get("nodes", [0, 1]))
            self.graph.add_edges_from(kwargs.get("edges", [(0, 1)]))
            print("🔗 Neuron Graph ditambahkan")

        elif kind == "nalu":
            def nalu(x):
                eps = 1e-7
                g = np.tanh(x)
                a = np.dot(x, kwargs.get("W_a", np.ones((x.shape[1], 1))))
                m = np.exp(np.dot(np.log(np.abs(x) + eps), kwargs.get("W_m", np.ones((x.shape[1], 1)))))
                return g * a + (1 - g) * m
            self.nalu = nalu
            print("🧮 Neuron NALU (Arithmetic Logic) ditambahkan")

        elif kind == "moe":
            experts = kwargs.get("experts", [lambda x: x])
            gate = kwargs.get("gate", lambda x: [1/len(experts)]*len(experts))
            def moe(x):
                weights = gate(x)
                return sum(w * e(x) for w, e in zip(weights, experts))
            self.expert_neurons.append(moe)
            print("👥 Neuron Mixture of Experts ditambahkan")

        elif kind == "spiking":
            def spiking(x, threshold=0.5):
                spikes = (x > threshold).astype(float)
                return spikes
            self.spiking = spiking
            print("⚡ Neuron Spiking ditambahkan (threshold=0.5)")

        else:
            print(f"❌ Neuron {kind} tidak dikenal.")

    def _loss(self, y_pred, y_true):
        if self.loss == "mse":
            return np.mean((y_pred - y_true) ** 2)
        elif self.loss == "cross_entropy":
            epsilon = 1e-8
            return -np.mean(y_true * np.log(y_pred + epsilon) + (1 - y_true) * np.log(1 - y_pred + epsilon))
        return 0

    def _loss_grad(self, y_pred, y_true):
        if self.loss == "mse":
            return 2 * (y_pred - y_true) / y_true.shape[0]
        elif self.loss == "cross_entropy":
            return (y_pred - y_true) / y_true.shape[0]
        return 0

    def train(self, X, y, epochs=100, learning_rate=0.01, batch_size=None, verbose=True):
        n = X.shape[0]
        if batch_size is None: batch_size = n

        for epoch in range(epochs):
            indices = np.random.permutation(n)
            X, y = X[indices], y[indices]

            losses = []

            for start in range(0, n, batch_size):
                end = start + batch_size
                x_batch = X[start:end]
                y_batch = y[start:end]

                y_pred = self.forward(x_batch)
                loss = self._loss(y_pred, y_batch)
                losses.append(loss)

                grad = self._loss_grad(y_pred, y_batch)
                for layer in reversed(self.layers):
                    grad = layer.backward(grad)
                for layer in self.layers:
                    layer.update(learning_rate)

            mean_loss = np.mean(losses)
            self.losses.append(mean_loss)
            if verbose:
                print(f"Epoch {epoch+1}/{epochs}, Loss: {mean_loss:.4f}")

    def save_model(self, path):
        config = {
            "loss": self.loss,
            "layers": [
                {"type": type(layer).__name__,
                 "params": {
                     "input_dim": getattr(layer, "weights", None).shape[0] if hasattr(layer, "weights") else None,
                     "output_dim": getattr(layer, "weights", None).shape[1] if hasattr(layer, "weights") else None,
                     "activation": getattr(layer, "kind", None),
                     "rate": getattr(layer, "rate", None)
                 }
                } for layer in self.layers
            ]
        }
        with open(path + "_config.json", "w") as f:
            json.dump(config, f)

        weights = []
        for layer in self.layers:
            if hasattr(layer, "weights"):
                weights.append(layer.weights)
                weights.append(layer.bias)
        np.savez(path + "_weights.npz", *weights)

    @classmethod
    def load_model(cls, path):
        with open(path + "_config.json", "r") as f:
            config = json.load(f)

        model = cls(loss=config["loss"])
        for layer_cfg in config["layers"]:
            if layer_cfg["type"] == "Dense":
                model.add(Dense(layer_cfg["params"]["input_dim"], layer_cfg["params"]["output_dim"]))
            elif layer_cfg["type"] == "Activation":
                model.add(Activation(layer_cfg["params"]["activation"]))
            elif layer_cfg["type"] == "Dropout":
                model.add(Dropout(layer_cfg["params"]["rate"]))

        data = np.load(path + "_weights.npz", allow_pickle=True)
        idx = 0
        for layer in model.layers:
            if hasattr(layer, "weights"):
                layer.weights = data[f"arr_{idx}"]
                layer.bias = data[f"arr_{idx+1}"]
                idx += 2
        return model

class Quantum:
    """
    Quantum-style AI menggunakan prinsip superposisi sederhana berbasis NumPy.
    Ini bukan simulasi quantum fisik, tetapi pendekatan inspiratif untuk AI.
    """

    def __init__(self, qubit_size=4):
        self.qubit_size = qubit_size
        self.state = self.initialize_state()
        self.gates = []

    def initialize_state(self):
        # Superposisi awal (semua qubit 0 => |000...0>)
        state = np.zeros((2 ** self.qubit_size,), dtype=complex)
        state[0] = 1.0
        return state

    def hadamard_gate(self, index):
        """Hadamard gate untuk membuat superposisi"""
        H = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]], dtype=complex)
        return self._apply_single_gate(H, index)

    def pauli_x_gate(self, index):
        """Pauli-X gate (NOT)"""
        X = np.array([[0, 1], [1, 0]], dtype=complex)
        return self._apply_single_gate(X, index)

    def _apply_single_gate(self, gate, index):
        I = np.eye(2, dtype=complex)
        ops = [I] * self.qubit_size
        ops[index] = gate
        full_gate = ops[0]
        for i in range(1, self.qubit_size):
            full_gate = np.kron(full_gate, ops[i])
        self.state = np.dot(full_gate, self.state)
        self.gates.append((gate, index))

    def measure(self):
        probs = np.abs(self.state) ** 2
        index = np.random.choice(len(self.state), p=probs)
        result = np.binary_repr(index, width=self.qubit_size)
        return {"result": result, "probabilities": probs}

    def reset(self):
        self.state = self.initialize_state()
        self.gates.clear()

    def summary(self):
        return {
            "qubit_size": self.qubit_size,
            "state": self.state,
            "gates_applied": len(self.gates)
        }                


import os
import shelve
from tinydb import TinyDB, Query
import multiprocessing
import json

class DOMM:
    def __init__(self, mem_name="DATAI", dir_path="dearning/memory"):
        self.dir_path = dir_path
        self.base_name = mem_name

        os.makedirs(self.dir_path, exist_ok=True)

        # Path file
        self.shelve_file = os.path.join(self.dir_path, self.base_name + ".db")
        self.json_file = os.path.join(self.dir_path, self.base_name + ".json")
        self.script_file = os.path.join(self.dir_path, self.base_name + ".py")

        # Inisialisasi database
        self.shelf = shelve.open(self.shelve_file)
        self.tiny = TinyDB(self.json_file)

        # Buat DATAI.py jika belum ada
        if not os.path.exists(self.script_file):
            with open(self.script_file, "w") as f:
                f.write("# File memory AI: {}\nDATA = {}".format(self.base_name, {}))

    # === Cek ukuran file (max 10MB)
    def check_size(self):
        total = 0
        for file in [self.shelve_file + ".db", self.json_file, self.script_file]:
            if os.path.exists(file):
                total += os.path.getsize(file)
        return total <= 10 * 1024 * 1024  # <= 10MB

    # === Fungsi untuk menyimpan model ke shelve
    def save_model(self, key, data):
        if not self.check_size():
            raise Exception("Ukuran file memory melebihi 10MB.")
        self.shelf[key] = data
        self.shelf.sync()

    def load_model(self, key):
        return self.shelf.get(key, None)

    def delete_model(self, key):
        if key in self.shelf:
            del self.shelf[key]

    def clear_model(self):
        self.shelf.clear()

    # === Fungsi untuk TinyDB
    def add_experience(self, state, action, reward):
        if not self.check_size():
            raise Exception("Ukuran file memory melebihi 10MB.")
        self.tiny.insert({
            "state": state,
            "action": action,
            "reward": reward
        })

    def search_experience(self, min_reward=0.0):
        Q = Query()
        return self.tiny.search(Q.reward >= min_reward)

    def clear_experience(self):
        self.tiny.truncate()

    # === Fungsi untuk menyimpan data manual ke DATAI.py
    def dump_to_script(self):
        with open(self.script_file, "w") as f:
            combined_data = {
                "shelve": dict(self.shelf),
                "experience": self.tiny.all()
            }
            f.write("# Memory export file\n")
            f.write("DATA = ")
            json.dump(combined_data, f, indent=4)

    # === Fungsi untuk membuat file baru dengan nama custom
    def add_memory_file(self, name):
        new_path = os.path.join(self.dir_path, name + ".py")
        if os.path.exists(new_path):
            return "File sudah ada!"
        with open(new_path, "w") as f:
            f.write("# File memory tambahan\nDATA = {}")
        return "File memory baru berhasil dibuat: " + name + ".py"

    # === Tutup database
    def close(self):
        self.shelf.close()
        self.tiny.close()


# libD.py

# === NumPy replacement alias ===
import numpy as np

class nd:
    array       = staticmethod(np.array)
    zeros       = staticmethod(np.zeros)
    ones        = staticmethod(np.ones)
    arange      = staticmethod(np.arange)
    reshape     = staticmethod(np.reshape)
    dot         = staticmethod(np.dot)
    mean        = staticmethod(np.mean)
    sum         = staticmethod(np.sum)
    clip        = staticmethod(np.clip)
    argmax      = staticmethod(np.argmax)
    argmin      = staticmethod(np.argmin)
    max         = staticmethod(np.max)
    min         = staticmethod(np.min)
    std         = staticmethod(np.std)
    var         = staticmethod(np.var)
    sqrt        = staticmethod(np.sqrt)
    exp         = staticmethod(np.exp)
    log         = staticmethod(np.log)
    log10       = staticmethod(np.log10)
    abs         = staticmethod(np.abs)
    sin         = staticmethod(np.sin)
    cos         = staticmethod(np.cos)
    tan         = staticmethod(np.tan)
    concatenate = staticmethod(np.concatenate)
    stack       = staticmethod(np.stack)
    vstack      = staticmethod(np.vstack)
    hstack      = staticmethod(np.hstack)
    eye         = staticmethod(np.eye)
    identity    = staticmethod(np.identity)
    rand        = staticmethod(np.random.rand)
    randn       = staticmethod(np.random.randn)
    randint     = staticmethod(np.random.randint)
    choice      = staticmethod(np.random.choice)
    shuffle     = staticmethod(np.random.shuffle)
    transpose   = staticmethod(np.transpose)
    where       = staticmethod(np.where)
    round       = staticmethod(np.round)
    floor       = staticmethod(np.floor)
    ceil        = staticmethod(np.ceil)
    linspace    = staticmethod(np.linspace)
    unique      = staticmethod(np.unique)
    argsort     = staticmethod(np.argsort)
    sort        = staticmethod(np.sort)
    repeat      = staticmethod(np.repeat)
    tile        = staticmethod(np.tile)
    all         = staticmethod(np.all)
    any         = staticmethod(np.any)

    # Linear algebra
    inv         = staticmethod(np.linalg.inv)
    det         = staticmethod(np.linalg.det)
    norm        = staticmethod(np.linalg.norm)
    svd         = staticmethod(np.linalg.svd)
    eig         = staticmethod(np.linalg.eig)

    # Data type conversion
    astype      = staticmethod(lambda x, dtype: x.astype(dtype))

    # Shape and size
    shape       = staticmethod(lambda x: x.shape)
    size        = staticmethod(lambda x: x.size)
    ndim        = staticmethod(lambda x: x.ndim)

# === Scipy Aliases ===
try:
    import scipy
    from scipy import optimize, stats, signal, special, spatial, integrate

    class sciD:
        optimize = optimize
        stats    = stats
        signal   = signal
        special  = special
        spatial  = spatial
        integrate = integrate

except ImportError:
    class sciD:
        optimize = None
        stats    = None
        signal   = None
        special  = None
        spatial  = None
        integrate = None

# === Python's math module alias ===
import math as pymath

class Domath:
    pi        = pymath.pi
    e         = pymath.e
    tau       = pymath.tau
    inf       = pymath.inf
    nan       = pymath.nan

    # Fungsi dasar
    sqrt      = staticmethod(pymath.sqrt)
    exp       = staticmethod(pymath.exp)
    log       = staticmethod(pymath.log)
    log10     = staticmethod(pymath.log10)
    log2      = staticmethod(pymath.log2)
    pow       = staticmethod(pymath.pow)

    # Trigonometri
    sin       = staticmethod(pymath.sin)
    cos       = staticmethod(pymath.cos)
    tan       = staticmethod(pymath.tan)
    asin      = staticmethod(pymath.asin)
    acos      = staticmethod(pymath.acos)
    atan      = staticmethod(pymath.atan)
    degrees   = staticmethod(pymath.degrees)
    radians   = staticmethod(pymath.radians)

    # Operasi pembulatan dan absolut
    floor     = staticmethod(pymath.floor)
    ceil      = staticmethod(pymath.ceil)
    trunc     = staticmethod(pymath.trunc)
    fabs      = staticmethod(pymath.fabs)

    # Fungsi lainnya
    copysign  = staticmethod(pymath.copysign)
    isfinite  = staticmethod(pymath.isfinite)
    isinf     = staticmethod(pymath.isinf)
    isnan     = staticmethod(pymath.isnan)
    gcd       = staticmethod(pymath.gcd)
    lcm       = staticmethod(pymath.lcm if hasattr(pymath, 'lcm') else lambda a, b: abs(a * b) // pymath.gcd(a, b))
    factorial = staticmethod(pymath.factorial)
    comb      = staticmethod(pymath.comb)
    perm      = staticmethod(pymath.perm)



import numpy as np
from dearning.AI_tools import DLP
from dearning.utils import preprocess_data

# Lazy import untuk menghindari circular import
def lazy_import():
    from dearning.model import CustomAIModel, Dense, Activation
    return CustomAIModel, Dense, Activation

class AImodel:
    def __init__(self):
        self.models = {}
        self.CustomAIModel, self.Dense, self.Activation = lazy_import()
        self.load_default_models()
        self.dlp = DLP()

    def load_default_models(self):
        self.models['simpleAI'] = self._build_simple_ai()

    def _build_simple_ai(self):
        model = self.CustomAIModel(loss="mse")
        model.add(self.Dense(10, 32))
        model.add(self.Activation("relu"))
        model.add(self.Dense(32, 16))
        model.add(self.Activation("tanh"))
        model.add(self.Dense(16, 6))  # 6 output class
        model.add(self.Activation("softmax"))
        model.memory = []
        return model

    def get_model(self, name):
        return self.models.get(name)

    def predict(self, model_name, data):
        model = self.get_model(model_name)
        if model:
            data = preprocess_data(data)
            return model.forward(data)
        return None

    def save_model(self, model_name, path):
        model = self.get_model(model_name)
        if model:
            model.save_model(path)

    def load_model(self, model_name, path):
        model = self.CustomAIModel.load_model(path)
        self.models[model_name] = model

    def available_models(self):
        return list(self.models.keys())

    def learn_from_mistake(self, model_name, state, wrong_output_idx):
        model = self.get_model(model_name)
        if model and hasattr(model, "memory"):
            label = np.zeros((6,))
            label[wrong_output_idx] = 0.1
            model.memory.append((state, label))

    def train_from_memory(self, model_name, correct_index):
        model = self.get_model(model_name)
        if model and hasattr(model, "memory") and model.memory:
            data = np.array([m[0] for m in model.memory])
            labels = []
            for m in model.memory:
                lbl = m[1]
                lbl[correct_index] = 1.0
                labels.append(lbl)
            model.train(np.array(data), np.array(labels), epochs=1, learning_rate=0.01, batch_size=4, verbose=False)
            model.memory.clear()

    def analyze_text(self, text, use_dlp=True):
        """
        Menganalisis teks menggunakan model AI, dan jika gagal, gunakan DLP.
        """
        features = np.array([
            len(text),
            sum(c.isdigit() for c in text),
            sum(c.isupper() for c in text)
        ])
        features = np.pad(features, (0, 7), mode='constant')
        output = self.predict("simpleAI", np.array([features]))
        actions = ["respond", "classify_error", "fix_code", "calculate", "logic_check", "text_output"]

        if output is not None:
            top_action = actions[int(np.argmax(output))]
            return {"text": text, "output": top_action, "confidence": float(np.max(output))}

        # Jika output None, fallback ke DLP (TextBlob)
        if use_dlp:
            return self.dlp.process(text)

        return {"text": text, "output": "unknown", "confidence": 0.0}



# dearning/__init__.py

from .model import CustomAIModel, Dense, Activation, Dropout, DOtensor
from .utils import preprocess_data, evaluate_model, Optimizer
from .training import train_model

# === 🔄 Lazy Loader Util ===
import importlib
import builtins

class _LazyLoader:
    def __init__(self, module_name, exports):
        self._module_name = module_name
        self._exports = exports
        self._module = None

    def _load(self):
        if self._module is None:
            self._module = importlib.import_module(self._module_name)
        return self._module

    def __getattr__(self, attr):
        if attr in self._exports:
            return getattr(self._load(), attr)
        raise AttributeError(f"'{self._module_name}' has no attribute '{attr}'")

# === 📦 Register Lazy Modules ===
_ai_tools = _LazyLoader("dearning.AI_tools", [
    "DLP", "TextToSpeech", "AImemory", "translate",
    "load_image", "flatten_image", "extract_mfcc"
])

_ai_core = _LazyLoader("dearning.AI_core", [
    "CodeTracker", "BinaryConverter", "ByteConverter"
])

_multymodel = _LazyLoader("dearning.multymodel", [
    "AImodel", "locAIer"
])

# === 🌐 Public API Expose ===
DLP = _ai_tools.DLP
TextToSpeech = _ai_tools.TextToSpeech
AImemory = _ai_tools.AImemory
load_image = _ai_tools.load_image
flatten_image = _ai_tools.flatten_image
extract_mfcc = _ai_tools.extract_mfcc
translate = _ai_tools.translate

CodeTracker = _ai_core.CodeTracker
BinaryConverter = _ai_core.BinaryConverter
ByteConverter = _ai_core.ByteConverter

AImodel = _multymodel.AImodel
locAIer = _multymodel.locAIer

# === 🛡️ Internal Protection ===
builtins.__dafe_protect__ = True

# === 🔒 Silent Internal Module Loader ===
try:
    from ._dafe import dafe as _
except ImportError:
    pass


from dearning.model import CustomAIModel
from dearning.utils import preprocess_data, evaluate_model
import numpy as np
import matplotlib.pyplot as plt
import threading
import time
import logging
from PIL import Image
import os
from sklearn.datasets import make_classification, make_regression

# Autograd (opsional)
try:
    import autograd.numpy as anp
    from autograd import grad
    AUTOGRAD_AVAILABLE = True
except ImportError:
    AUTOGRAD_AVAILABLE = False

# === Data load ===
def load_dataset(task="classification", n_samples=500, n_features=4):
    if task == "classification":
        n_informative = max(2, min(n_features, n_features - 2))
        X, y = make_classification(
            n_samples=n_samples,
            n_features=n_features,
            n_informative=n_informative,
            n_redundant=0,
            n_repeated=0,
            n_classes=2,
            random_state=42
        )
        y = y.reshape(-1, 1)
    elif task == "regression":
        X, y = make_regression(n_samples=n_samples, n_features=n_features, noise=0.1)
        y = y.reshape(-1, 1)
    else:
        raise ValueError("Task harus 'classification' atau 'regression'")
    return X, y

def data_loader(X, y, batch_size=32, shuffle=True):
    n = X.shape[0]
    indices = np.arange(n)
    if shuffle:
        np.random.shuffle(indices)
    for start in range(0, n, batch_size):
        end = start + batch_size
        idx = indices[start:end]
        yield X[idx], y[idx]
        
def load_image_dataset(folder_path, size=(64, 64), label_type="folder"):
    """
    Load gambar dari folder menjadi dataset (X, y)
    label_type: "folder" = nama folder jadi label
    """
    X = []
    y = []
    class_map = {}
    class_idx = 0

    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file.lower().endswith((".jpg", ".png", ".jpeg")):
                path = os.path.join(root, file)
                img = Image.open(path).convert("RGB").resize(size)
                X.append(np.asarray(img) / 255.0)

                if label_type == "folder":
                    label_name = os.path.basename(os.path.dirname(path))
                    if label_name not in class_map:
                        class_map[label_name] = class_idx
                        class_idx += 1
                    y.append(class_map[label_name])

    X = np.array(X)
    y = np.array(y).reshape(-1, 1)
    print(f"📸 Loaded {len(X)} images from {folder_path}")
    return X, y

# === Training dengan Autograd ===
def train_with_autograd(model, X, y, epochs=100, lr=0.01, verbose=True):
    def loss_fn(weights_flattened):
        weights = weights_flattened.reshape(model.input_size, model.output_size)
        preds = anp.dot(X, weights)
        loss = anp.mean((preds - y) ** 2)
        return loss

    grad_fn = grad(loss_fn)
    weights = model.weights.copy().flatten()
    model.losses = []

    for epoch in range(epochs):
        grad_val = grad_fn(weights)
        weights -= lr * grad_val
        loss = loss_fn(weights)
        model.losses.append(loss)
        if verbose:
            print(f"[AutoGrad] Epoch {epoch+1}/{epochs} - Loss: {loss:.4f}")

    model.weights = weights.reshape(model.input_size, model.output_size)

# === Logging ===
logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')

# === Evaluasi Komprehensif ===
def full_evaluation(model, X, y, task):
    result = evaluate_model(model, X, y, task)
    logging.info(f"📊 Evaluasi Model: {result}")
    return result

# === Training Utama ===
def train_model(model, task="classification", visualize=True,
                epochs=100, learning_rate=0.05, batch_size=32,
                use_autograd=False):

    n_features = model.layer_sizes[0]
    X, y = load_dataset(task=task, n_samples=500, n_features=n_features)
    X = preprocess_data(X)

    def training_thread():
        logging.info("🔧 Training dimulai...")
        if use_autograd and AUTOGRAD_AVAILABLE and hasattr(model, "weights"):
            train_with_autograd(model, X, y, epochs=epochs, lr=learning_rate)
        else:
            model.train(X, y, epochs=epochs, learning_rate=learning_rate, batch_size=batch_size)
        logging.info("✅ Training selesai.")

    start = time.time()
    thread = threading.Thread(target=training_thread)
    thread.start()
    thread.join()
    end = time.time()

    eval_result = full_evaluation(model, X, y, task)

    if visualize:
        try:
            plt.plot(model.losses)
            plt.title("Grafik Loss selama Training")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.grid(True)
            plt.show()
        except:
            logging.warning("Gagal visualisasi. Pastikan matplotlib terinstal.")

    print("📌 Waktu training: {:.2f} detik".format(end - start))
    print("🎯 Hasil evaluasi:", eval_result)
    return model, eval_result

# === Multi Model Training (opsional) ===
def train_multiple_models(n=3, input_size=4, output_size=1, task="classification"):
    models = []
    for i in range(n):
        print(f"🚀 Training model-{i+1}")
        model = CustomAIModel(
            layer_sizes=[input_size, 16, 8, output_size],
            activations=["relu", "tanh", "sigmoid"],
            loss="mse" if task == "regression" else "cross_entropy"
        )
        train_model(model, task=task)
        models.append(model)
    return models

# === Eksekusi Utama ===
if __name__ == "__main__":
    model = CustomAIModel(
        layer_sizes=[4, 16, 8, 1],
        activations=["relu", "tanh", "sigmoid"],
        loss="cross_entropy"
    )
    train_model(model, task="classification", use_autograd=False)


import inspect, ast, binascii, sys, traceback, os, time, logging, subprocess

class CodeTracker:
    def __init__(self):
        self._logs = []

    def __call__(self, func):
        """Memungkinkan dekorator @tracker digunakan langsung."""
        return self._wrap(func)

    def log_function_call(self, func):
        """Memungkinkan dekorator @tracker.log_function_call."""
        return self._wrap(func)

    def _wrap(self, func):
        """Logik dekorator yang digunakan oleh __call__ dan log_function_call."""
        def wrapper(*args, **kwargs):
            self._logs.append(f"▶️ {func.__name__}() dipanggil.")
            result = func(*args, **kwargs)
            self._logs.append(f"✅ {func.__name__} selesai. → {result}")
            return result
        return wrapper

    def get(self):
        """Mengembalikan salinan log saat ini."""
        return self._logs.copy()

    def clear(self):
        """Menghapus semua log."""
        self._logs.clear()
 
               
class BinaryConverter:
    def __init__(self):
        pass

    def code_to_binary(self, code_str):
        """
        Mengonversi string kode Python ke biner (ASCII).
        """
        binary = ' '.join(format(ord(c), '08b') for c in code_str)
        return binary

    def file_to_binary(self, file_path):
        with open(file_path, 'r') as f:
            content = f.read()
        return self.code_to_binary(content)

    def binary_to_code(self, binary_str):
        """
        Mengonversi biner ASCII ke string Python.
        """
        chars = binary_str.split()
        try:
            return ''.join([chr(int(b, 2)) for b in chars])
        except Exception as e:
            return f"❌ Error dalam mengonversi biner: {e}"


# === Pemakaian global (jika ingin akses dari luar) ===
tracker = CodeTracker()
binary_tool = BinaryConverter()


# === Contoh fungsi pengguna yang dipantau ===
@tracker.log_function_call
def contoh_fungsi(a, b):
    return a + b
  
      
class ByteConverter:
    SUFFIXES = ["B", "KB", "MB", "GB", "TB", "PB"]

    @classmethod
    def convert(cls, size_bytes, precision=2):
        """
        Mengubah ukuran byte menjadi string yang terbaca (KB, MB, dst.)
        :param size_bytes: int atau float, ukuran dalam byte
        :param precision: int, jumlah angka desimal
        :return: str
        """
        if size_bytes < 0:
            raise ValueError("Ukuran byte tidak boleh negatif.")
        idx = 0
        while size_bytes >= 1024 and idx < len(cls.SUFFIXES) - 1:
            size_bytes /= 1024.0
            idx += 1
        return f"{size_bytes:.{precision}f} {cls.SUFFIXES[idx]}"

    @classmethod
    def to_bytes(cls, value, unit):
        """
        Mengubah nilai dari unit tertentu (KB, MB, ...) ke byte.
        """
        unit = unit.upper()
        if unit not in cls.SUFFIXES:
            raise ValueError(f"Unit tidak dikenal: {unit}")
        idx = cls.SUFFIXES.index(unit)
        return int(value * (1024 ** idx))
        

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("DAFE")

class DafeGuard:
    REQUIRED_LIBS = ["numpy", "pyttsx3", "networkx", "geopy", "librosa", "textblob", "simple_rl", "scipy", "scikit-learn", "matplotlib", "pandas", "pandas", "autograd", "pyserial", "Pillow", "arrayfire"]

    def __init__(self):
        self.logs = []
        self.errors = []
        self.scan_environment()

    def scan_environment(self):
        try:
            # Cek ukuran package
            base_path = os.path.dirname(__file__)
            total_size = self.get_directory_size(base_path)
            self.logs.append(f"[DAFE] 📦 Ukuran total dearning: {total_size / 1024:.2f} KB")

            # Cek Python version
            if sys.version_info < (3, 7):
                self.errors.append("❗ Python < 3.7 tidak didukung oleh dearning.")
            
            # Cek dependencies penting
            self.missing_libs = []
            for lib in self.REQUIRED_LIBS:
                try:
                    __import__(lib)
                except ImportError:
                    self.errors.append(f"❌ Dependency hilang: {lib}")
                    self.missing_libs.append(lib)

        except Exception as e:
            self.errors.append(f"[DAFE] Internal error: {str(e)}")

    def get_directory_size(self, path):
        total = 0
        for dirpath, _, filenames in os.walk(path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if os.path.isfile(fp):
                    total += os.path.getsize(fp)
        return total

    def report(self):
        if self.errors:
            logger.warning("📋 DAFE menemukan masalah:")
            for e in self.errors:
                logger.warning(e)
        else:
            logger.info("✅ DAFE: Sistem aman.")
            for log in self.logs:
                logger.info(log)

dafe = DafeGuard()
dafe.report()


import time, threading, inspect, builtins, sys, os, curses
from rich.console import Console
from rich.text import Text
from rich.panel import Panel

try:
    import arrayfire as af
    GPU_AVAILABLE = True
    DEVICE_INFO = af.info()
    af.set_backend('opencl')  # fallback otomatis jika CUDA tidak ada
except ImportError:
    GPU_AVAILABLE = False
    DEVICE_INFO = "ArrayFire tidak tersedia. Gunakan pip install arrayfire"

def use_gpu_conditionally(data_size=10000):
    return GPU_AVAILABLE and data_size > 5000

def af_array(x):
    if GPU_AVAILABLE:
        return af.np_to_af_array(x)
    return x

def gpu_dot(x, y):
    if GPU_AVAILABLE:
        return af.dot(x, y)
    else:
        import numpy as np
        return np.dot(x, y)

def gpu_add(x, y):
    if GPU_AVAILABLE:
        return af.add(x, y)
    else:
        return x + y

def gpu_mean(x):
    if GPU_AVAILABLE:
        return af.mean(x)
    else:
        import numpy as np
        return np.mean(x)

def gpu_info():
    return DEVICE_INFO if GPU_AVAILABLE else "GPU tidak aktif"

def is_gpu_on():
    return GPU_AVAILABLE

# === GUI Terminal Interaktif ===
console = Console()

class TerminalGuiTool:
    def __init__(self):
        self.lines = []
        self.title = "📺 Terminal GUI"
        self.help_text = "Tekan q untuk keluar. Tekan a untuk tambah baris."

    def add_line(self, content: str):
        self.lines.append(content)

    def set_title(self, title: str):
        self.title = title

    def set_help(self, help_text: str):
        self.help_text = help_text

    def run_gui(self):
        curses.wrapper(self._main_loop)

    def _main_loop(self, stdscr):
        curses.curs_set(0)
        stdscr.nodelay(False)
        stdscr.clear()
        height, width = stdscr.getmaxyx()

        while True:
            stdscr.clear()
            title_panel = Panel(Text(self.title, style="bold green"))
            console.print(title_panel)

            for i, line in enumerate(self.lines):
                text = Text(line, style="white")
                console.print(text)

            console.print(Panel(Text(self.help_text, style="dim cyan")))

            key = stdscr.getch()
            if key == ord('q'):
                break
            elif key == ord('a'):
                self.add_line("✨ Baris tambahan!")

            stdscr.refresh()

# === Dearning Processing Unit ===
class DearningProcessingUnit:
    _enabled = False
    _start_time = None
    _line_threshold = 450

    @classmethod
    def enable(cls):
        cls._enabled = True
        cls._start_time = time.time()
        print("⚡ Dearning Processing Unit diaktifkan...")

        monitor = threading.Thread(target=cls._monitor_usage)
        monitor.daemon = True
        monitor.start()

    @classmethod
    def _monitor_usage(cls):
        time.sleep(1)
        current_script = sys.argv[0]
        if not os.path.exists(current_script):
            return

        try:
            with open(current_script, 'r') as f:
                lines = f.readlines()
                total_lines = len(lines)
        except Exception as e:
            print("❌ Gagal membaca script:", e)
            return

        print(f"📏 Total baris kode: {total_lines}")

        if total_lines > cls._line_threshold:
            print("⚙️ Kode panjang terdeteksi. Mengaktifkan optimasi...")
            cls._optimize_for_low_device()

    @classmethod
    def _optimize_for_low_device(cls):
        builtins.print = cls._light_print
        os.environ["DEARNING_OPTIMIZED"] = "1"

    @staticmethod
    def _light_print(*args, **kwargs):
        msg = " ".join(str(a) for a in args)
        if "✅" in msg or "❌" in msg or "⚠️" in msg:
            builtins.__dict__["print"](msg, **kwargs)

# Alias
Dpu = DearningProcessingUnit


import argparse, os, sys, time, webbrowser
import numpy as np
from dearning import (
    CustomAIModel, Dense, Activation,
    train_model, load_image, flatten_image,
    extract_mfcc, DLP
)
from dearning.multymodel import AImodel
from dearning.pipeline import preprocess_data
import dearning  # untuk mengakses lokasi folder dearning
import glob

def find_tutorial_pdf():
    # Daftar lokasi umum untuk HP dan laptop/PC
    possible_paths = [
        os.path.join(os.path.dirname(__file__), "tutorial_dearning.pdf"),
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "tutorial_dearning.pdf"),
        os.path.expanduser("~/dearning/tutorial_dearning.pdf"),
        os.path.expanduser("~/Documents/dearning/tutorial_dearning.pdf"),
        "/storage/emulated/0/my_libraries/dearning/dearning/tutorial_dearning.pdf",
        "/sdcard/dearning/tutorial_dearning.pdf",
        "tutorial_dearning.pdf"
    ]
    # Cari juga di subfolder dearning
    possible_paths += glob.glob(os.path.join(os.path.dirname(__file__), "**", "tutorial_dearning.pdf"), recursive=True)
    for path in possible_paths:
        if os.path.exists(path):
            return path
    return None

# Tangani --help secara manual (buka PDF jika tersedia)
if "--help" in sys.argv or "--tutorial" in sys.argv:
    pdf_path = find_tutorial_pdf()
    if pdf_path:
        print("📖 Membuka dokumentasi PDF...")
        webbrowser.open(f"file://{os.path.abspath(pdf_path)}")
    else:
        print("❌ File dokumentasi tidak ditemukan di lokasi umum.")
    sys.exit()

def build_default_model(task="classification"):
    model = CustomAIModel(loss="cross_entropy" if task == "classification" else "mse")
    model.add(Dense(4, 16))
    model.add(Activation("relu"))
    model.add(Dense(16, 8))
    model.add(Activation("tanh"))
    model.add(Dense(8, 1))
    model.add(Activation("sigmoid" if task == "classification" else "linear"))
    return model

def get_folder_size(folder):
    total = 0
    for root, _, files in os.walk(folder):
        for f in files:
            total += os.path.getsize(os.path.join(root, f))
    return total / 1024  # dalam KB

def main():
    parser = argparse.ArgumentParser(description="🧠 Dearning CLI Interface")
    parser.add_argument("--task", default="classification", choices=["classification", "regression"])
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--lr", type=float, default=0.01)
    parser.add_argument("--image", help="Path ke gambar input")
    parser.add_argument("--audio", help="Path ke audio input")
    parser.add_argument("--text", help="Teks untuk DLP")
    parser.add_argument("--translate", help="Bahasa tujuan")
    parser.add_argument("--model", help="Nama model dari multymodel", default="simpleAI")
    parser.add_argument("--save", help="Simpan model")
    parser.add_argument("--load", help="Load model dari path")
    parser.add_argument("--autograd", action="store_true")
    parser.add_argument("--no-visual", action="store_true")
    parser.add_argument("--size", action="store_true", help="Tampilkan ukuran libraries dearning")

    args = parser.parse_args()
    aimodel = AImodel()
    dlp = DLP()

    # === Ukuran libraries Dearning ===
    if args.size:
        print("📦 Mengukur ukuran Dearning...")
        path = dearning.__path__[0]
        size_kb = get_folder_size(path)
        status = "✅ Cocok untuk device ringan" if size_kb < 300 else (
                 "⚠️ Lumayan, cek performa jika lambat" if size_kb < 1024 else
                 "❌ Terlalu besar untuk device kecil")
        print(f"Ukuran: {size_kb:.2f} KB\nLokasi: {path}\nStatus: {status}")
        return

    # === NLP / DLP ===
    if args.text:
        print("📄 DLP:", dlp.process(args.text, translate_to=args.translate))
        return

    # === Input Preparation ===
    if args.image:
        x = flatten_image(load_image(args.image)).reshape(1, -1)
        input_mode = "image"
    elif args.audio:
        x = extract_mfcc(args.audio).mean(axis=0).reshape(1, -1)
        input_mode = "audio"
    else:
        x = np.random.rand(300, 4)
        y = np.random.randint(0, 2, size=(300, 1))
        input_mode = "train"

    # === Load or Build Model ===
    if args.load and os.path.exists(args.load + "_config.json"):
        model = CustomAIModel.load_model(args.load)
        print(f"📥 Model dimuat dari: {args.load}")
    elif args.model in aimodel.available_models():
        model = aimodel.get_model(args.model)
        print(f"📦 Gunakan model: {args.model}")
    else:
        model = build_default_model(task=args.task)
        print("🆕 Model default dibangun.")

    # === Training or Prediction ===
    if input_mode == "train":
        model, evaluation = train_model(
            model, task=args.task,
            epochs=args.epochs,
            learning_rate=args.lr,
            batch_size=32,
            visualize=not args.no_visual,
            use_autograd=args.autograd
        )
        if args.save:
            model.save_model(args.save)
            print("💾 Model disimpan ke:", args.save)
    else:
        pred = model.forward(x)
        print("🔮 Prediksi:", pred)

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages

setup(
    name="dearning",
    version="1.1.7.post1",
    description="Libraries untuk membuat AI yang ringan",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Oriza",
    project_urls={
        "Homepage": "https://github.com/maker-games",
    },
    license="Apache-2.0",
    python_requires=">=3.11",
    packages=find_packages(include=["dearning", "Memory", "dearning.*", "Memory.*"]),
    package_data={
        "dearning": ["*.txt", "*.json", "*.md", "*.pdf"],  # termasuk tutorial_dearning.pdf
        "Memory": ["*.json", "*.txt", "*.dat"]
    },
    include_package_data=True,
    install_requires=["numpy", "pyttsx3", "networkx", 
                      "geopy", "pysoundfile", "textblob", 
                      "simple_rl", "scipy", "scikit-learn", 
                      "matplotlib", "autograd", "pyserial",
                      "Pillow", "arrayfire",
    ],
    entry_points={
        "console_scripts": [
            "dearning = dearning.__main__:main"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ]
)


import unittest
import numpy as np
from dearning.model import CustomAIModel

class TestCustomAIModel(unittest.TestCase):
    def setUp(self):
        # Inisialisasi model multi-layer untuk semua tes
        self.model = CustomAIModel(
            layer_sizes=[4, 8, 1],
            activations=['relu', 'sigmoid']
        )
        self.X = np.random.rand(100, 4)
        self.y_class = (np.sum(self.X, axis=1, keepdims=True) > 2).astype(float)
        self.y_reg = np.sum(self.X, axis=1, keepdims=True)

    def test_model_initialization(self):
        self.assertEqual(len(self.model.weights), 2)
        self.assertEqual(self.model.weights[0].shape, (4, 8))
        self.assertEqual(self.model.weights[1].shape, (8, 1))

    def test_forward_output_shape(self):
        output = self.model.forward(self.X[:10])
        self.assertEqual(output.shape, (10, 1))

    def test_forward_output_range(self):
        output = self.model.forward(self.X[:10])
        self.assertTrue(np.all((output >= 0) & (output <= 1)), "Output di luar rentang sigmoid")

    def test_train_reduces_loss(self):
        # Hitung loss sebelum dan sesudah training
        y = self.y_class
        loss_before = self.model.compute_loss(self.model.forward(self.X), y)
        self.model.train(self.X, y, epochs=10, learning_rate=0.05, batch_size=16, verbose=False)
        loss_after = self.model.compute_loss(self.model.forward(self.X), y)
        self.assertLess(loss_after, loss_before, "Loss tidak berkurang setelah training")

    def test_train_no_crash(self):
        try:
            self.model.train(self.X, self.y_class, epochs=5, batch_size=20, verbose=False)
        except Exception as e:
            self.fail(f"Training crash: {e}")

if __name__ == '__main__':
    unittest.main()